// ********************************************************************
// Demoprogramm
// EA T123A-I2C Display ber I2C (3x12 Zeichen) ansteuern
// LCD-Controller: PCF2116C
// Stand: 28.04.2008, Autor: Matthias Kahnt, Url: pro-51.eltra-tec.de
// SDCC-Version: M-IDE-51/SDCC 2.8.0
// AT89C51ED2: --code-loc 0x0000 --iram -size 256 --xram-loc 0x0000 --xram-size 0x0700
// AT89S8253:  --code-loc 0x0000 --iram-size 256
// ********************************************************************
// Fr 8051-Microcontroller, z.B. AT89C51ED2 oder AT89S8253
// Frequenz: 11.0592 MHz
// Code-Speicherbedarf: ca. 1kByte, keine XRAM-Verwendung
// Das Programm verwendet nur 8051/52 Standard-SFRs und wurde
// mit dem AT89C51ED2 getestet.
// ********************************************************************
// Die Programmfunktionen drfen fr private Nutzung verwendet und
// verndert werden. Fr gewerbliche Nutzung ist die Zustimmung
// der Authoren erforderlich. Die Nutzung erfolgt auf eigene Gefahr.
// Fr eventuell entstehende Schden wird keine Haftung bernommen.
// ********************************************************************

#include "at89c51ed2.h"   // Register des AT89C51ED2
//#include "at89s8253.h"   // Register des AT89S8253
#include <string.h>      // Stingfunktionen

// Typendefinition
typedef unsigned char byte;
typedef unsigned int  word;
typedef unsigned long dword;

// Definebereich
#define LED_D5      P0_7

#define SDA         P1_0
#define SCL         P1_1
#define	ACK	    0
#define	NACK	    1
#define ADDDIS      0x74

#define SPALTEN_MAX  12
#define ZEILEN_MAX   3

static bit i2c_error = 0;

// ********************************************************************
// Verzgerungsschleife fr kurze Zeit
// Zeit: Wartezeit in [zeit]
//       1 = 20us
//       2 = 28us
//       3 = 36us u.s.w. (gilt fr 11.0952MHz)
//     255 = ca. 2ms
// ********************************************************************
void delay_us(byte zeit) {
byte zaehler;
  for ( zaehler = zeit; zaehler; zaehler-- ) {
   _asm nop; _endasm;
  }
}

// ********************************************************************
// Verzgerungsschleife fr lange Zeit
// Zeit: Wartezeit in [zeit]
//       1 = ca. 10ms
//       2 = ca. 20ms
//       3 = ca. 30ms u.s.w.
//     255 = ca. 2,5s (gilt fr 11.0952MHz)
// ********************************************************************
void delay(byte zeit) {
byte zaehler;
  for (zaehler = zeit; zaehler; zaehler--) {
    delay_us(255);  // dauert ca. 2ms
    delay_us(255);  // dauert ca. 2ms
    delay_us(255);  // dauert ca. 2ms
    delay_us(255);  // dauert ca. 2ms
    delay_us(255);  // dauert ca. 2ms
  }
}

// ********************************************************************
// I2C-Bus
// Startsequenz
// ********************************************************************
void I2C_start(void) {
  SDA = 1; SDA = 1; SDA = 1; SDA = 1;
  SCL = 1; SCL = 1; SCL = 1; SCL = 1;
  SDA = 0;
}

// ********************************************************************
// I2C-Bus
// Stoppsequenz
// ********************************************************************
void I2C_stop(void) {
  SDA = 0; SDA = 0; SDA = 0; SDA = 0;
  SCL = 1; SCL = 1; SCL = 1; SCL = 1;
  SDA = 1;
}

// ********************************************************************
// I2C-Bus
// Write: 1 Byte schreiben
// bergabe: senddat - zu bertragendes Zeichen
// ********************************************************************
void I2C_write(byte senddat) {
byte i;
  SCL = 0;
  for (i = 1; i <= 8; i++) { // 8 Bit seriell senden
   SDA = senddat >> 7;
   SCL = 1; SCL = 1;
   senddat <<= 1;
   SCL = 0;
  }
  SDA = 1; // Portpin SDA fr Empfang das ACK-Bits vorbereiten
  SCL = 0;
  SCL = 1; SCL = 1; SCL = 1; SCL = 1;
  if (SDA) i2c_error = 1; // ACK erwartet, aber nicht erhalten
  SCL = 0;
}

// ********************************************************************
// I2C-Bus
// Read: 1 Byte schreiben
// bergabe: acknow - ACK (weiteres Zeichen anfordern)
//                    NACK (letztes Zeichen)
// ********************************************************************
byte I2C_read(byte acknow) {
byte empfdat, i;
  empfdat = 0;
  SDA = 1; // Portpin SDA fr Empfang vorbereiten
  SCL = 0;
  for (i = 1; i <= 8; i++) { // 8 Bit seriell empfangen
    SCL = 1; SCL = 1; SCL = 1; SCL = 1;
    empfdat <<= 1;
    empfdat |= SDA;
    SCL = 0;
  }
  SDA = acknow; // ACK-Bit senden
  SCL = 0;
  SCL = 1; SCL = 1; SCL = 1; SCL = 1;
  if (acknow == NACK) SDA = 1; // nchster Zyklus ist Reset
  SCL = 0; SCL = 0;
  SDA = 1;          /* Release the sda line */
  return empfdat;
}

// ***********************************************************************
// Schreiben eines Zeichens an das LCD-Modul
// bergabe: lcd_byte : Auszugebendes Zeichen/Steuerzeichen
//           lcd_mode : 0 - Daten
//                      1 - Steuerzeichen
// ***********************************************************************
void write_lcd(byte lcd_byte, bit lcd_mode) {
  delay_us(1); // kurze Pause (mind. 20us), da Busy nicht ausgewertet wird
  I2C_start();          // Startsequenz
  I2C_write(ADDDIS);    // Baustein fr schreiben adressieren
  if (lcd_mode) I2C_write(0x00); // Controlbyte Co=0 RS=0 Steuerzeichen
     else I2C_write(0x40);       // Controlbyte Co=0 RS=1 Daten
  I2C_write(lcd_byte);  // Datenbyte
  I2C_stop();           // Stoppsequenz
} 

// ***********************************************************************
// Lschen des LCD-Display
//         Kursor auf Position 0,0
// ***********************************************************************
void clear_lcd(void) {
  write_lcd(0x01,1); // Clear LCD
  delay(2);  // Clear-Befehl dauert ca. 1.1ms
} 

// ***********************************************************************
// Ausgabe eines ASCII-Zeichen positioniert auf dem LCD-Modul
// bergabe: lcd_x : Spalte (0...SPALTEN_MAX-1)
//           lcd_y : Zeile  (0...ZEILEN_MAX-1)
//           lcd_ascii : ASCII-Zeichen
// *********************************************************************** 
void write_byte_lcd(byte lcd_x, byte lcd_y, byte lcd_ascii) {
byte lcd_offset = 0;
  if (lcd_x > (SPALTEN_MAX - 1)) lcd_x = 0;
  if (lcd_y > (ZEILEN_MAX  - 1)) lcd_y = 0;
  switch (lcd_y) {
    case 0:  lcd_offset = 0x00; break;  // Zeile 1
    case 1:  lcd_offset = 0x20; break;  // Zeile 2 
    case 2:  lcd_offset = 0x40; break;  // Zeile 3 
  };
  write_lcd(0x80 + lcd_x + lcd_offset,1); // Kursorposition setzen
  write_lcd(lcd_ascii | 0x80,0);          // Ausgabe des ASCII-Zeichen an der Kursorposition
}

// ***********************************************************************
// Ausgabe einer Zeichenkette positioniert auf dem LCD-Modul
// bergabe: lcd_x : Spalte (0...SPALTEN_MAX-1)
//           lcd_y : Zeile  (0...ZEILEN_MAX-1)
//           lcd_zeichen : Adresse der auszugebenden format.  Zeichenkette
//           clr_line    : Lschen bis Zeilenende
//                         1 - Lschen
//                         0 - kein Lschen
// ***********************************************************************
void printf_lcd(byte lcd_x, byte lcd_y, byte *lcd_zeichen, bit clr_line) {
byte lcd_i;
byte lcd_offset = 0;
  if (lcd_x > (SPALTEN_MAX - 1)) lcd_x = 0;
  if (lcd_y > (ZEILEN_MAX  - 1)) lcd_y = 0;
  switch (lcd_y) {
    case 0:  lcd_offset = 0x00; break;  // Zeile 1
    case 1:  lcd_offset = 0x20; break;  // Zeile 2 
    case 2:  lcd_offset = 0x40; break;  // Zeile 3 
  }
  write_lcd(0x80 + lcd_x + lcd_offset,1);  // Kursorposition setzen
  // Ausgabe der Zeichenkette ab der Kursorposition
  lcd_offset = strlen(lcd_zeichen);        // Lnge der Zeichenkette
  if (lcd_offset > SPALTEN_MAX) lcd_offset = SPALTEN_MAX;
  for(lcd_i = lcd_offset ; lcd_i ; lcd_i--) {
    write_lcd((*lcd_zeichen) | 0x80,0); lcd_zeichen++;
  }
  if (clr_line) {
    // Lschen bis Zeilenende
    for(lcd_i = SPALTEN_MAX - lcd_offset - lcd_x ; lcd_i ; lcd_i--)
      write_lcd(' ',0);
  }
}

// ***********************************************************************
// Blinkposition auf dem LCD-Displays setzen / Blinken ein/ausschalten
// bergabe: lcd_x : Spalte (0...SPALTEN_MAX-1)
//           lcd_y : Zeile  (0...ZEILEN_MAX-1)
//           lcd_blink :   0 - Blinken/Cursor aus
//                         1 - Blinken an
//                         2 - Cursor  an
// ***********************************************************************
void blink_lcd(byte lcd_x, byte lcd_y, byte lcd_blink) {
byte lcd_offset = 0;
  write_lcd(0x0C,1); // KURSOR ausschalten
  if (lcd_x > (SPALTEN_MAX - 1)) lcd_x = 0;
  if (lcd_y > (ZEILEN_MAX  - 1)) lcd_y = 0;
  switch (lcd_y) {
    case 0:  lcd_offset = 0x00; break;  // Zeile 1
    case 1:  lcd_offset = 0x20; break;  // Zeile 2 
    case 2:  lcd_offset = 0x40; break;  // Zeile 3 
  };
  write_lcd(0x80 + lcd_x + lcd_offset,1); // Blinkposition setzen
  if (lcd_blink == 1) write_lcd(0x0D,1);  // Blinken ein
  if (lcd_blink == 2) write_lcd(0x0E,1);  // Cursor ein
}

// ***********************************************************************
// Grundinitialisierung des LCD-Moduls
// ***********************************************************************
void init_lcd(void) {
  I2C_start();          // Startsequenz
  I2C_write(ADDDIS);    // Display fr schreiben adressieren
  I2C_write(0x00);      // Co=0 Controlbyte RS=0 (Steuerbyte)
  I2C_write(0x3E);      // Funktion Set    0011 1110 DL=1 N=1 M=1 G=1
  I2C_write(0x08);      // Display Control 0000 1000 Disp=0 Cur=0 Blink=0
  I2C_write(0x06);      // Entry Mode Set  0000 0110 I/D=1 S=0
  I2C_write(0x0C);      // Display Control 0000 1100 Disp=1
  I2C_stop();           // Stoppsequenz
} 

// ********************************************************************
// Grund-Initialisierung
// ********************************************************************
void init_system (void) {
  LED_D5 = 0; // Pro-51 LED einschalten
}

// ********************************************************************
// Hauptprogramm
// ********************************************************************
void main (void) {
  init_system();
  init_lcd();
  clear_lcd();  // LCD lschen
  printf_lcd(0,0,"PRO-51 3x12",1); // Einschaltmeldung
  printf_lcd(0,1,"EA T123A-I2C",1);  
  printf_lcd(0,2,"PCF2116C",1);  
  write_byte_lcd(11,2,'.');  
  blink_lcd(11,2,1);
  while(1);
}
